(function() {
  /*
    File storage for Node Index Module
  
    This software is licensed under the MIT License.
  
    Copyright Fedor Indutny, 2011.
  
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to permit
    persons to whom the Software is furnished to do so, subject to the
    following conditions:
  
    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.
  
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
  var Buffer, DEFAULT_OPTIONS, Storage, fs, isPosition, packer, path, step, utils;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  utils = require('./utils');
  step = require('step');
  fs = require('fs');
  path = require('path');
  packer = {
    pack: function(data) {
      var result, size;
      data = JSON.stringify(data);
      size = Buffer.byteLength(data);
      result = new Buffer(size + 4);
      result[0] = (size >> 24) & 0xff;
      result[1] = (size >> 16) & 0xff;
      result[2] = (size >> 8) & 0xff;
      result[3] = size & 0xff;
      result.write(data, 4);
      return result;
    },
    unpack: function(data) {
      var size;
      size = (data[0] << 24) + (data[1] << 16) + (data[2] << 8) + data[3];
      return JSON.parse((data.slice(4, 4 + size)).toString());
    }
  };
  Buffer = require('buffer').Buffer;
  /*
    Default storage options
  */
  DEFAULT_OPTIONS = {
    filename: '',
    padding: 8,
    rootSize: 64,
    partitionSize: 1024 * 1024 * 1024,
    flushMinBytes: 4 * 1024 * 1024
  };
  /*
    Class @constructor
  */
  Storage = exports.Storage = function(options, callback) {
    options = utils.merge(DEFAULT_OPTIONS, options);
    if (!options.filename) {
      return callback('Filename is required');
    }
    this.filename = options.filename, this.padding = options.padding, this.partitionSize = options.partitionSize, this.rootSize = options.rootSize, this.flushMinBytes = options.flushMinBytes;
    this._init(callback);
    return this;
  };
  /*
    Initializes storage
    May be used not from constructor
  */
  Storage.prototype._init = function(callback) {
    this.files = [];
    this.filesOffset = 0;
    this.buffers = [];
    this.buffersBytes = 0;
    this.buffersHashmap = {};
    return this.checkCompaction(__bind(function(err) {
      if (err) {
        return callback(err);
      }
      return this.openFile(__bind(function(err) {
        if (err) {
          return callback(err);
        }
        if (this.files.length <= 0) {
          return this.createFile(callback);
        } else {
          return callback(null, this);
        }
      }, this));
    }, this));
  };
  /*
    @constructor wrapper
  */
  exports.createStorage = function(options, callback) {
    return new Storage(options, callback);
  };
  /*
    pos = [
      start-offset,
      length
      file-index or undefined
    ]
  */
  Storage.prototype.isPosition = isPosition = function(pos) {
    return (pos != null) && Array.isArray(pos) && pos.length === 3 && true;
  };
  /*
    Adds index to the end of file
    and return next (unopened) filename
  */
  Storage.prototype.nextFilename = function(i, filename) {
    var index;
    index = i || this.files.length;
    index -= this.filesOffset;
    filename = filename || this.filename;
    if (index > 0) {
      return filename + '.' + index;
    } else {
      return filename;
    }
  };
  /*
    Check if interrupted compaction is in place
  */
  Storage.prototype.checkCompaction = function(callback) {
    var filename, nextFilename;
    filename = this.filename;
    nextFilename = this.nextFilename.bind(this);
    return path.exists(filename, __bind(function(exists) {
      return path.exists(filename + '.compact', __bind(function(compactExists) {
        if (exists) {
          if (compactExists) {
            this.iterateFiles(filename + '.compact', function(err, i) {
              var compacted;
              if (err) {
                throw err;
              }
              if (i !== null) {
                compacted = nextFilename(i, filename + '.compact');
                return fs.unlink(compacted, this.parallel());
              } else {
                return this.parallel()(null);
              }
            });
          }
          return callback(null);
        } else {
          if (compactExists) {
            return this.iterateFiles(filename + '.compact', function(err, i) {
              var compacted, notcompacted, _callback;
              if (err) {
                throw err;
              }
              if (i !== null) {
                compacted = nextFilename(i, filename + '.compact');
                notcompacted = nextFilename(i, filename);
                _callback = this.parallel();
                return fs.unlink(notcompacted, function(err) {
                  if (err && (!err.code || err.code !== 'ENOENT')) {
                    return _callback(err);
                  } else {
                    return fs.rename(compacted, notcompacted, _callback);
                  }
                });
              } else {
                return callback(null);
              }
            });
          } else {
            return callback(null);
          }
        }
      }, this));
    }, this));
  };
  /*
    Iterate through files in descending order
    filename.N
    filename.N - 1
    ...
    filename
  */
  Storage.prototype.iterateFiles = function(filename, callback) {
    var files, iterate, next;
    files = [];
    next = __bind(function() {
      var _filename;
      _filename = this.nextFilename(files.length, filename);
      return step(function() {
        var _callback;
        _callback = this.parallel();
        path.exists(_filename, function(exists) {
          return _callback(null, exists);
        });
      }, function(err, exists) {
        if (!exists) {
          process.nextTick(iterate);
          return;
        }
        files.push(files.length);
        return process.nextTick(next);
      });
    }, this);
    iterate = function() {
      var file;
      file = files.pop();
      if (file == null) {
        file = null;
      }
      return step(function() {
        return this.parallel()(null, file);
      }, callback, function(err) {
        if (err) {
          return callback(err);
        } else if (file !== null) {
          return process.nextTick(iterate);
        }
      });
    };
    return next();
  };
  /*
    Add index to the end of filename,
    open it if exists and store size
  */
  Storage.prototype.openFile = function(callback) {
    var file, filename, files, padding, that;
    that = this;
    padding = this.padding;
    files = this.files;
    filename = this.nextFilename();
    file = {};
    return step(function() {
      var _callback;
      _callback = this.parallel();
      path.exists(filename, function(exists) {
        return _callback(null, exists);
      });
    }, function(err, exists) {
      if (err) {
        this.paralell()(err);
        return;
      }
      if (!exists) {
        this.parallel()(null);
        return;
      }
      fs.open(filename, 'a+', 0666, this.parallel());
      fs.stat(filename, this.parallel());
    }, function(err, fd, stat) {
      var index, paddBuff;
      if (err) {
        throw err;
      }
      if (!(fd && stat)) {
        this.parallel()(null);
        return;
      }
      index = files.length;
      file = {
        filename: filename,
        fd: fd,
        size: stat.size,
        index: index
      };
      if (file.size % padding) {
        paddBuff = new Buffer(padding - file.size % padding);
        fs.write(fd, paddBuff, 0, paddBuff.length, null, this.parallel());
      } else {
        this.parallel()(null, 0);
      }
      this.parallel()(null, file);
      files.push(file);
    }, function(err, bytesWritten, file) {
      if (err) {
        throw err;
      }
      if (bytesWritten != null) {
        if (file.size % padding) {
          if (bytesWritten !== padding - file.size % padding) {
            this.parallel()('Can\'t add padding to db file');
            return;
          }
          file.size += padding - file.size % padding;
        }
        return that.openFile(this.parallel());
      } else {
        return this.parallel()(null);
      }
    }, callback);
  };
  /*
    Create file
  */
  Storage.prototype.createFile = function(writeRoot, callback) {
    var filename;
    filename = this.nextFilename();
    if (callback == null) {
      callback = writeRoot;
      writeRoot = true;
    }
    return fs.open(filename, 'w+', 0666, __bind(function(err, fd) {
      var file;
      if (err) {
        return callback(err);
      }
      file = {
        filename: filename,
        fd: fd,
        size: 0,
        index: this.files.length - this.filesOffset
      };
      this.files.push(file);
      if (!writeRoot) {
        return callback(null, this);
      }
      this.root_pos = ['0', '2', '0'];
      this.root_pos_data = null;
      return this.write([], __bind(function(err, pos) {
        if (err) {
          return callback(err);
        }
        return this.writeRoot(pos, __bind(function(err) {
          if (err) {
            return callback(err);
          }
          return callback(null, this);
        }, this));
      }, this));
    }, this));
  };
  /*
    Read data from position
  */
  Storage.prototype.read = function(pos, callback, raw) {
    var buff, cached, f, file, l, s;
    if (!isPosition(pos)) {
      return callback('pos should be a valid position (read)');
    }
    s = pos[0];
    l = pos[1];
    f = pos[2];
    file = this.files[f || 0];
    buff = new Buffer(l);
    if (!file) {
      return callback('pos is incorrect');
    }
    cached = this.buffersHashmap[pos.join('-')];
    if (cached) {
      try {
        if (!raw) {
          cached = packer.unpack(cached);
        }
        callback(null, cached);
        return;
      } catch (e) {

      }
    }
    return fs.read(file.fd, buff, 0, l, s, function(err, bytesRead) {
      if (err) {
        return callback(err);
      }
      if (bytesRead !== l) {
        return callback('Read less bytes than expected');
      }
      if (!raw) {
        try {
          buff = packer.unpack(buff);
          err = null;
        } catch (e) {
          err = 'Data is not a valid json';
        }
      }
      return callback(err, buff);
    });
  };
  /*
    Write data and return position
  */
  Storage.prototype.write = function(data, callback, raw) {
    if (raw && !Buffer.isBuffer(data)) {
      return callback('In raw mode data should be a buffer');
    }
    if (!raw) {
      data = new Buffer(packer.pack(data));
    }
    return this._fsWrite(data, callback);
  };
  /*
    Read root page
  */
  Storage.prototype.readRoot = function(callback) {
    var cache_callback;
    if (this.root_pos_data) {
      callback(null, this.root_pos_data);
      return;
    }
    cache_callback = __bind(function(err, data) {
      if (err) {
        return callback(err);
      }
      this.root_pos_data = data;
      callback(null, data);
    }, this);
    if (this.root_pos) {
      this.read(this.root_pos, cache_callback);
      return;
    }
    return this.readRootPos(__bind(function(err, pos) {
      if (err) {
        return callback(err);
      }
      return this.read(pos, cache_callback);
    }, this));
  };
  /*
    Find last root in files and return it to callback
  
    it will be synchronous for now
    TODO: make it asynchronous
  */
  Storage.prototype.readRootPos = function(callback) {
    var checkHash, iterate;
    iterate = __bind(function(index, callback) {
      var buff, bytesRead, data, file, offset, root;
      file = this.files[index];
      if (!file) {
        return callback('root not found');
      }
      buff = new Buffer(this.rootSize);
      offset = file.size - (file.size % this.padding) - this.rootSize + this.padding;
      while ((offset -= this.padding) >= 0) {
        bytesRead = fs.readSync(file.fd, buff, 0, this.rootSize, offset);
        if (bytesRead !== this.rootSize) {
          offset = -1;
          break;
        }
        if (data = checkHash(buff)) {
          root = data;
          try {
            root = packer.unpack(root);
          } catch (e) {
            offset = -1;
            break;
          }
          return callback(null, root);
        }
      }
      return process.nextTick(function() {
        return iterate(index - 1, callback);
      });
    }, this);
    checkHash = function(buff) {
      var hash, rest;
      hash = buff.slice(0, utils.hash.len).toString();
      rest = buff.slice(utils.hash.len);
      if (hash === utils.hash(rest)) {
        return rest;
      }
    };
    return iterate(this.files.length - 1, callback);
  };
  /*
    Write root page
  */
  Storage.prototype.writeRoot = function(root_pos, callback) {
    var buff, hash, _root_pos;
    if (!isPosition(root_pos)) {
      return callback('pos should be a valid position (writeRoot)');
    }
    _root_pos = packer.pack(root_pos);
    buff = new Buffer(this.rootSize);
    _root_pos.copy(buff, utils.hash.len);
    hash = utils.hash(buff.slice(utils.hash.len));
    buff.write(hash, 0, 'binary');
    return this._fsBuffer(buff, true, __bind(function(err) {
      if (err) {
        return callback(err);
      }
      this.root_pos = root_pos;
      this.root_pos_data = null;
      return callback(null);
    }, this));
  };
  /*
    Low-level write
  
    buff - is Buffer
  
  */
  Storage.prototype._fsWrite = function(buff, callback) {
    var file, pos;
    file = this.currentFile();
    pos = [file.size, buff.length, file.index];
    file.size += buff.length;
    this.buffersHashmap[pos.join('-')] = buff;
    return this._fsBuffer(buff, false, function(err) {
      return callback(err, pos);
    });
  };
  /*
    Low-level bufferization
  */
  Storage.prototype._fsBuffer = function(buff, isRoot, callback) {
    var buffLen, delta, fd, file;
    file = this.currentFile();
    fd = file.fd;
    buffLen = buff.length;
    if (isRoot) {
      if (file.size % this.padding) {
        delta = this.padding - (file.size % this.padding);
        buffLen += delta;
        this.buffers.push(new Buffer(delta));
      }
      file.size += buffLen;
    }
    this.buffers.push(buff);
    if ((this.buffersBytes += buffLen) < this.flushMinBytes) {
      callback(null);
      return;
    }
    return this._fsFlush(callback);
  };
  /*
    Flush buffered data to disk
  */
  Storage.prototype._fsFlush = function(callback, compacting) {
    var buff, buffLen, buffers, fd, file, i, offset, _ref;
    file = this.currentFile();
    fd = file.fd;
    buffLen = this.buffersBytes;
    buff = new Buffer(buffLen);
    offset = 0;
    buffers = this.buffers;
    for (i = 0, _ref = buffers.length; 0 <= _ref ? i < _ref : i > _ref; 0 <= _ref ? i++ : i--) {
      buffers[i].copy(buff, offset);
      offset += buffers[i].length;
    }
    this.buffers = [];
    this.buffersBytes = 0;
    this.buffersHashmap = {};
    return fs.write(fd, buff, 0, buffLen, null, __bind(function(err, bytesWritten) {
      if (err || (bytesWritten !== buffLen)) {
        this._fsCheckSize(function(err2) {
          return callback(err2 || err || 'Written less bytes than expected');
        });
        return;
      }
      if (file.size >= this.partitionSize && !compacting) {
        return this.createFile(false, callback);
      } else {
        return callback(null);
      }
    }, this));
  };
  /*
    Recheck current file's length
  */
  Storage.prototype._fsCheckSize = function(callback) {
    var file, filename;
    file = this.currentFile();
    filename = this.nextFilename(this.file.index);
    return fs.stat(filename, __bind(function(err, stat) {
      if (err) {
        return callback(err);
      }
      file.size = stat.size;
      return callback(null);
    }, this));
  };
  /*
    Current file
  */
  Storage.prototype.currentFile = function() {
    return this.files[this.files.length - 1];
  };
  /*
    Close all fds
  */
  Storage.prototype.close = function(callback) {
    var files;
    files = this.files;
    return this._fsFlush(function(err) {
      if (err) {
        return callback(err);
      }
      return step(function() {
        var group, i, _ref, _results;
        group = this.group();
        _results = [];
        for (i = 0, _ref = files.length; 0 <= _ref ? i < _ref : i > _ref; 0 <= _ref ? i++ : i--) {
          _results.push(files[i] ? fs.close(files[i].fd, group()) : void 0);
        }
        return _results;
      }, callback);
    });
  };
  /*
    Compaction flow actions
  */
  Storage.prototype.beforeCompact = function(callback) {
    this.filename += '.compact';
    this.filesOffset = this.files.length;
    return this._fsFlush(__bind(function(err) {
      if (err) {
        return callback(err);
      }
      return this.createFile(false, callback);
    }, this), true);
  };
  Storage.prototype.afterCompact = function(callback) {
    var files, filesOffset, that;
    that = this;
    filesOffset = this.filesOffset;
    files = this.files;
    this.filename = this.filename.replace(/\.compact$/, '');
    return this._fsFlush(__bind(function(err) {
      if (err) {
        return callback(err);
      }
      return step(function() {
        var i, _ref;
        for (i = 0, _ref = files.length; 0 <= _ref ? i < _ref : i > _ref; 0 <= _ref ? i++ : i--) {
          fs.close(files[i].fd, this.parallel());
        }
      }, function(err) {
        var i;
        if (err) {
          throw err;
        }
        for (i = 0; 0 <= filesOffset ? i < filesOffset : i > filesOffset; 0 <= filesOffset ? i++ : i--) {
          fs.unlink(files[i].filename, this.parallel());
        }
      }, function(err) {
        var compactedCount, fnsQueue, i, _i, _ref, _results;
        if (err) {
          throw err;
        }
        fnsQueue = [];
        compactedCount = files.length - filesOffset;
        (function() {
          _results = [];
          for (var _i = 0; 0 <= compactedCount ? _i < compactedCount : _i > compactedCount; 0 <= compactedCount ? _i++ : _i--){ _results.push(_i); }
          return _results;
        }).apply(this).forEach(function(i) {
          var compactedName, normalName;
          compactedName = files[i + filesOffset].filename;
          normalName = files[i].filename;
          files[i] = files[i + filesOffset];
          files[i].filename = normalName;
          return fnsQueue.unshift(function(err) {
            if (err) {
              throw err;
            }
            fs.rename(compactedName, normalName, this.parallel());
          });
        });
        fnsQueue.push(this.parallel());
        step.apply(null, fnsQueue);
        for (i = compactedCount, _ref = files.length; compactedCount <= _ref ? i < _ref : i > _ref; compactedCount <= _ref ? i++ : i--) {
          files.pop();
        }
        that.filesOffset = 0;
      }, function(err) {
        var _i, _ref, _results;
        if (err) {
          throw err;
        }
        (function() {
          _results = [];
          for (var _i = 0, _ref = files.length; 0 <= _ref ? _i < _ref : _i > _ref; 0 <= _ref ? _i++ : _i--){ _results.push(_i); }
          return _results;
        }).apply(this).forEach(__bind(function(i) {
          var file, fn;
          file = files[i];
          fn = this.parallel();
          return fs.open(file.filename, 'a+', 0666, function(err, fd) {
            file.fd = fd;
            return fn(err, file);
          });
        }, this));
      }, function(err) {
        if (err) {
          step(function() {
            var i, _ref;
            for (i = 0, _ref = files.length; 0 <= _ref ? i < _ref : i > _ref; 0 <= _ref ? i++ : i--) {
              fs.close(files[i].fd, this.parallel());
            }
          }, function() {
            that._init(this.parallel());
          }, this.parallel());
          return;
        }
        return null;
      }, callback);
    }, this));
  };
}).call(this);
